# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for apport.packaging_impl.apt_dpkg."""

import tempfile
import unittest
import unittest.mock
from unittest.mock import MagicMock

import apt

from apport.packaging_impl.apt_dpkg import (
    WITH_DEB822_SUPPORT,
    _parse_deb822_sources,
    _read_mirror_file,
    impl,
)


@unittest.mock.patch(
    "apport.packaging_impl.apt_dpkg.__AptDpkgPackageInfo.get_os_version",
    MagicMock(return_value=("Ubuntu", "22.04")),
)
class TestPackagingAptDpkg(unittest.TestCase):
    """Unit tests for apport.packaging_impl.apt_dpkg."""

    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_no_candidate(self, apt_cache_mock):
        """is_distro_package() for package that has no candidate."""
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package, installed=None, candidate=None
        )
        self.assertEqual(impl.is_distro_package("adduser"), False)
        getitem_mock.assert_called_once_with("adduser")

    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_no_installed_version(self, apt_cache_mock):
        """is_distro_package() for not installed package."""
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package,
            installed=MagicMock(spec=apt.package.Version, version=None),
        )
        self.assertEqual(impl.is_distro_package("7zip"), False)
        getitem_mock.assert_called_once_with("7zip")

    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_ppa(self, apt_cache_mock):
        """is_distro_package() for a PPA package."""
        version = MagicMock(
            spec=apt.package.Version,
            origins=[MagicMock(spec=apt.package.Origin, origin="LP-PPA")],
            version="0.5.0-0ppa1",
        )
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package, candidate=version, installed=version
        )
        self.assertEqual(impl.is_distro_package("bdebstrap"), False)
        getitem_mock.assert_called_once_with("bdebstrap")

    @unittest.mock.patch("os.path.exists")
    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_system_image(self, apt_cache_mock, exists_mock):
        """is_distro_package() for a system image without cache."""
        version = MagicMock(
            spec=apt.package.Version,
            origins=[MagicMock(spec=apt.package.Origin, origin="")],
            version="2.23.1-0ubuntu4",
        )
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package, candidate=version, installed=version
        )
        exists_mock.return_value = True
        self.assertEqual(impl.is_distro_package("apport"), True)
        getitem_mock.assert_called_once_with("apport")
        exists_mock.assert_called_once_with("/etc/system-image/channel.ini")

    @unittest.skipIf(not WITH_DEB822_SUPPORT, "no deb822 support")
    @unittest.mock.patch(
        "builtins.open",
        new_callable=unittest.mock.mock_open,
        read_data="""
# Some documentation in the beginning

Types: deb deb-src
URIs: http://example.com
Suites: foo foo-bar
Components: main


Types: deb
URIs: http://example2.com
Suites:
 baz
Components: main


""",
    )
    def test_parse_deb822_sources_extra_lines(self, mock_file):
        """Test _parse_deb822_sources with multiple lines separating blocks."""
        entries = _parse_deb822_sources("foo_bar.sources")
        mock_file.assert_called_with("foo_bar.sources", encoding="utf-8")
        self.assertEqual(len(entries), 2)
        self.assertEqual(entries[0].uris[0], "http://example.com")
        self.assertEqual(entries[1].suites[0], "baz")

    def test_read_mirror_file(self) -> None:
        """Test _read_mirror_file with config from GitHub CI."""
        with tempfile.NamedTemporaryFile("w") as mirror_file:
            mirror_file.write(
                "http://azure.archive.ubuntu.com/ubuntu/\tpriority:1\n"
                "http://archive.ubuntu.com/ubuntu/\tpriority:2\n"
                "http://security.ubuntu.com/ubuntu/\tpriority:3\n"
            )
            mirror_file.flush()
            mirrors = _read_mirror_file(f"mirror+file:{mirror_file.name}")
        self.assertEqual(
            mirrors,
            [
                "http://azure.archive.ubuntu.com/ubuntu/",
                "http://archive.ubuntu.com/ubuntu/",
                "http://security.ubuntu.com/ubuntu/",
            ],
        )



# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for data/apport."""

import datetime
import errno
import io
import os
import pathlib
import shutil
import signal
import sys
import tempfile
import time
import unittest
from unittest.mock import MagicMock

import apport.fileutils
import apport.user_group
from tests.helper import import_module_from_file
from tests.paths import get_data_directory

apport_binary = import_module_from_file(get_data_directory() / "apport")


class TestApport(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    """Unit tests for data/apport."""

    @classmethod
    def setUpClass(cls):
        cls.orig_report_dir = apport.fileutils.report_dir

    @classmethod
    def tearDownClass(cls):
        apport.fileutils.report_dir = cls.orig_report_dir

    def setUp(self):
        self.workdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.workdir)
        apport.fileutils.report_dir = os.path.join(self.workdir, "crash")
        self.report_dir = pathlib.Path(apport.fileutils.report_dir)

    @unittest.mock.patch("subprocess.run")
    def test_check_kernel_crash(self, run_mock):
        """Test found kernel crash dump."""
        self.report_dir.mkdir()
        vmcore = self.report_dir / "vmcore"
        vmcore.touch()
        apport_binary.check_kernel_crash()
        run_mock.assert_called_once_with(
            ["/usr/share/apport/kernel_crashdump"], check=False
        )

    def test_check_lock_not_writable(self) -> None:
        """Test check_lock() with not writable lock file."""
        os.environ["APPORT_LOCK_FILE"] = "/non-existing/apport.lock"
        try:
            with self.assertRaisesRegex(SystemExit, "^1$"):
                apport_binary.check_lock()
        finally:
            del os.environ["APPORT_LOCK_FILE"]

    @unittest.mock.patch("fcntl.lockf")
    def test_check_lock_taken(self, lockf_mock: MagicMock) -> None:
        """Test check_lock() with lock file taken by other process."""
        # Since this is a unit test, let the mock raise an exception instead
        # of letting the fcntl.lockf call run into the timeout signal.
        lockf_mock.side_effect = TimeoutError()
        with tempfile.NamedTemporaryFile("w") as lockfile:
            os.environ["APPORT_LOCK_FILE"] = lockfile.name
            try:
                with self.assertRaisesRegex(SystemExit, "^1$"):
                    apport_binary.check_lock()
            finally:
                del os.environ["APPORT_LOCK_FILE"]

    def test_refine_core_ulimit_huge(self) -> None:
        """Test refine_core_ulimit() with huge limit."""
        options = apport_binary.parse_arguments(
            ["-p", str(os.getpid()), "-c", str(pow(2, 64))]
        )
        self.assertEqual(apport_binary.refine_core_ulimit(options), -1)

    @unittest.mock.patch("os.isatty")
    def test_init_error_log_is_tty(self, isatty_mock: MagicMock) -> None:
        """Test init_error_log() doing nothing on a TTY."""
        isatty_mock.return_value = True
        stderr = sys.stderr
        apport_binary.init_error_log()
        # Check sys.stderr to be unchanged
        self.assertEqual(sys.stderr, stderr)
        isatty_mock.assert_called_once_with(2)

    def test_receive_arguments_via_socket_import_error(self) -> None:
        """Test receive_arguments_via_socket() fail to import systemd."""
        with (
            self.assertLogs(level="ERROR") as error_logs,
            unittest.mock.patch("builtins.__import__") as import_mock,
            self.assertRaisesRegex(SystemExit, "^0$"),
        ):
            import_mock.side_effect = ModuleNotFoundError("No module named 'systemd'")
            apport_binary.receive_arguments_via_socket()

        import_mock.assert_called_once()
        self.assertRegex(
            error_logs.output[0], "apport-forward.socket.*systemd python module"
        )

    def test_receive_arguments_via_socket_invalid_socket(self):
        """Test receive_arguments_via_socket with invalid socket."""
        try:
            # pylint: disable=import-outside-toplevel
            from systemd.daemon import listen_fds
        except ImportError:
            self.skipTest("systemd Python module not available")
        assert listen_fds
        self.assertNotIn("LISTEN_FDS", os.environ)
        with self.assertRaisesRegex(SystemExit, "^1$"):
            apport_binary.receive_arguments_via_socket()

    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    @unittest.mock.patch.object(
        apport_binary, "is_same_ns", MagicMock(return_value=False)
    )
    @unittest.mock.patch.object(apport_binary, "forward_crash_to_container")
    def test_main_forward_crash_to_container(self, forward_mock):
        """Test main() to forward crash to container."""
        args = ["-p", "12345", "-P", "67890"]
        self.assertEqual(apport_binary.main(args), 0)
        forward_mock.assert_called_once()

    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    @unittest.mock.patch.object(apport_binary, "start_apport")
    def test_main_start(self, start_mock):
        """Test calling apport with --start."""
        self.assertEqual(apport_binary.main(["--start"]), 0)
        start_mock.assert_called_once_with()

    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    @unittest.mock.patch.object(apport_binary, "stop_apport")
    def test_main_stop(self, stop_mock):
        """Test calling apport with --stop."""
        self.assertEqual(apport_binary.main(["--stop"]), 0)
        stop_mock.assert_called_once_with()

    def test_start(self):
        """Test starting Apport crash handler."""
        open_mock = unittest.mock.mock_open()
        with unittest.mock.patch("builtins.open", open_mock):
            apport_binary.start_apport()
        open_mock.assert_called_with(
            "/proc/sys/kernel/core_pipe_limit", "w", encoding="utf-8"
        )
        self.assertEqual(open_mock.call_count, 3)

    def test_consistency_checks_replaced_process(self):
        """Test consistency_checks() for a replaced crash process ID."""
        pid = os.getpid()
        options = apport_binary.parse_arguments(["-p", str(pid)])
        now = int(time.clock_gettime(time.CLOCK_BOOTTIME) * 100)
        crash_user = apport.user_group.get_process_user_and_group()
        with apport_binary.ProcPid(pid) as proc_pid:
            self.assertFalse(
                apport_binary.consistency_checks(options, now, proc_pid, crash_user)
            )

    def test_consistency_checks_mismatching_uid(self):
        """Test consistency_checks() for a mitmatching UID."""
        pid = os.getpid()
        crash_user = apport.user_group.get_process_user_and_group()
        options = apport_binary.parse_arguments(
            [
                "-p",
                str(pid),
                "-u",
                str(crash_user.uid + 1),
                "-g",
                str(crash_user.gid + 1),
            ]
        )
        with apport_binary.ProcPid(pid) as proc_pid:
            self.assertFalse(
                apport_binary.consistency_checks(options, 1, proc_pid, crash_user)
            )

    def test_stop(self):
        """Test stopping Apport crash handler."""
        open_mock = unittest.mock.mock_open()
        with unittest.mock.patch("builtins.open", open_mock):
            apport_binary.stop_apport()
        open_mock.assert_called_with(
            "/proc/sys/kernel/core_pattern", "w", encoding="utf-8"
        )
        self.assertEqual(open_mock.call_count, 3)

    @unittest.mock.patch("os.setresgid", MagicMock())
    @unittest.mock.patch("os.setresuid", MagicMock())
    @unittest.mock.patch.object(apport_binary, "_run_with_output_limit_and_timeout")
    @unittest.mock.patch("os.path.exists")
    def test_is_closing_session(
        self, path_exist_mock: MagicMock, run_mock: MagicMock
    ) -> None:
        """Test is_closing_session()."""
        path_exist_mock.return_value = True
        run_mock.return_value = (b"(false,)\n", b"some stderr output\n")
        crash_user = apport.user_group.UserGroupID(1337, 1337)
        with tempfile.TemporaryDirectory() as tmpdir:
            env = pathlib.Path(tmpdir) / "environ"
            env.write_text("DBUS_SESSION_BUS_ADDRESS=unix:path=/run/user/1337/bus\0")
            with apport_binary.ProcPid(12345, tmpdir) as proc_pid:
                self.assertEqual(
                    apport_binary.is_closing_session(proc_pid, crash_user), True
                )
        path_exist_mock.assert_called_once_with("/run/user/1337/bus")
        run_mock.assert_called_once()

    def test_is_closing_session_no_environ(self) -> None:
        """Test is_closing_session() with no DBUS_SESSION_BUS_ADDRESS."""
        crash_user = apport.user_group.UserGroupID(1337, 1337)
        with tempfile.TemporaryDirectory() as tmpdir:
            env = pathlib.Path(tmpdir) / "environ"
            env.write_text("DISPLAY=:0\0")
            with apport_binary.ProcPid(12345, tmpdir) as proc_pid:
                self.assertEqual(
                    apport_binary.is_closing_session(proc_pid, crash_user), False
                )

    def test_is_closing_session_no_determine_socket(self) -> None:
        """Test is_closing_session() cannot determine D-Bus socket."""
        crash_user = apport.user_group.UserGroupID(42, 42)
        with tempfile.TemporaryDirectory() as tmpdir:
            env = pathlib.Path(tmpdir) / "environ"
            env.write_text("DBUS_SESSION_BUS_ADDRESS=unix:/run/user/42/bus\0")
            with apport_binary.ProcPid(12345, tmpdir) as proc_pid:
                self.assertEqual(
                    apport_binary.is_closing_session(proc_pid, crash_user), False
                )

    def test_is_closing_session_socket_not_exists(self) -> None:
        """Test is_closing_session() where D-Bus socket does not exist."""
        assert not os.path.exists("/run/user/1337/bus")
        crash_user = apport.user_group.UserGroupID(1337, 1337)
        with tempfile.TemporaryDirectory() as tmpdir:
            env = pathlib.Path(tmpdir) / "environ"
            env.write_text("DBUS_SESSION_BUS_ADDRESS=unix:path=/run/user/1337/bus\0")
            with apport_binary.ProcPid(12345, tmpdir) as proc_pid:
                self.assertEqual(
                    apport_binary.is_closing_session(proc_pid, crash_user), False
                )

    @unittest.mock.patch("os.setresgid", MagicMock())
    @unittest.mock.patch("os.setresuid", MagicMock())
    @unittest.mock.patch.object(apport_binary, "_run_with_output_limit_and_timeout")
    @unittest.mock.patch("os.path.exists")
    def test_is_closing_session_gdbus_failure(
        self, path_exist_mock: MagicMock, run_mock: MagicMock
    ) -> None:
        """Test is_closing_session() with OSError from gdbus."""
        path_exist_mock.return_value = True
        run_mock.side_effect = OSError(
            errno.ENOENT, "No such file or directory: 'gdbus'"
        )
        crash_user = apport.user_group.UserGroupID(1337, 1337)
        with tempfile.TemporaryDirectory() as tmpdir:
            env = pathlib.Path(tmpdir) / "environ"
            env.write_text("DBUS_SESSION_BUS_ADDRESS=unix:path=/run/user/1337/bus\0")
            with apport_binary.ProcPid(12345, tmpdir) as proc_pid:
                self.assertEqual(
                    apport_binary.is_closing_session(proc_pid, crash_user), False
                )
        path_exist_mock.assert_called_once_with("/run/user/1337/bus")
        run_mock.assert_called_once()

    @unittest.mock.patch.object(apport_binary, "check_lock", MagicMock())
    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    def test_missing_proc_pid(self) -> None:
        """Test /proc/<pid> is already gone."""
        fake_pid = 2147483647
        assert not os.path.exists(f"/proc/{fake_pid}")
        with self.assertLogs(level="ERROR") as error_logs:
            with tempfile.NamedTemporaryFile() as stdin:
                stdin_mock = io.FileIO(stdin.name)
                with unittest.mock.patch("sys.stdin", return_value=stdin_mock):
                    self.assertEqual(apport_binary.main(["-p", str(fake_pid)]), 1)
        self.assertIn("/proc/2147483647 not found", error_logs.output[0])

    @unittest.mock.patch.object(apport_binary, "check_lock", MagicMock())
    @unittest.mock.patch.object(apport_binary, "consistency_checks", MagicMock())
    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    @unittest.mock.patch.object(
        apport_binary, "is_closing_session", MagicMock(return_value=False)
    )
    @unittest.mock.patch.object(apport_binary, "is_systemd_watchdog_restart")
    def test_main_ignore_watchdog_restart(
        self, is_systemd_watchdog_restart_mock: MagicMock
    ) -> None:
        """Test main() to ignore watchdog restarts.

        The underlying is_systemd_watchdog_restart() function is mocked,
        which should be tested in a separate unit test.
        """
        is_systemd_watchdog_restart_mock.return_value = True
        args = ["-p", str(os.getpid()), "-s", str(int(signal.SIGABRT)), "-c", "-1"]
        with self.assertLogs(level="ERROR") as error_logs:
            with tempfile.NamedTemporaryFile() as stdin:
                stdin_mock = io.TextIOWrapper(io.FileIO(stdin.name))
                with unittest.mock.patch("sys.stdin", stdin_mock):
                    exit_code = apport_binary.main(args)

        self.assertEqual(exit_code, 0)
        is_systemd_watchdog_restart_mock.assert_called_once()
        self.assertIn("Ignoring systemd watchdog restart", error_logs.output[0])

    @unittest.mock.patch.object(apport_binary, "init_error_log", MagicMock())
    def test_non_existing_systemd_coredump(self) -> None:
        """Test main() to print error if systemd-coredump cannot be found."""
        try:
            # pylint: disable-next=import-outside-toplevel
            import systemd.journal
        except ImportError as error:
            self.skipTest(f"{error.name} Python module not available")
        assert systemd.journal

        with self.assertLogs(level="ERROR") as error_logs:
            exit_code = apport_binary.main(["--from-systemd-coredump", "42-5846584-0"])

        self.assertEqual(exit_code, 1)
        self.assertIn(
            "No journal log for systemd unit"
            " systemd-coredump@42-5846584-0.service found.",
            error_logs.output[0],
        )

    def test_systemd_journal_import_error(self) -> None:
        """Test handling missing systemd.journal library correctly."""
        with (
            self.assertLogs(level="ERROR") as error_logs,
            unittest.mock.patch("builtins.__import__") as import_mock,
            self.assertRaisesRegex(SystemExit, "^1$"),
        ):
            import_mock.side_effect = ModuleNotFoundError(
                "No module named 'systemd.journal'"
            )
            apport_binary.process_crash_from_systemd_coredump("0-383264-0")

        import_mock.assert_called_once()
        self.assertRegex(error_logs.output[0], "Please install python3-systemd")

    @unittest.mock.patch.object(apport_binary, "process_crash")
    @unittest.mock.patch.object(apport_binary, "get_systemd_coredump")
    def test_reading_core_from_journal_log(
        self, get_systemd_coredump_mock: MagicMock, process_crash_mock: MagicMock
    ) -> None:
        """Test _user_can_read_coredump via process_crash_from_systemd_coredump

        The core dump can be provided via the journal log. In this case
        _user_can_read_coredump should behave correctly.
        """
        get_systemd_coredump_mock.return_value = {
            "COREDUMP": b"(\xb5/\xfd$\x0ca\x00\x00mocked core\nG\xfe\xe0\x10",
            "COREDUMP_CMDLINE": "python3",
            "COREDUMP_CWD": "/",
            "COREDUMP_ENVIRON": "SHELL=/bin/bash\n",
            "COREDUMP_EXE": sys.executable,
            "COREDUMP_GID": 0,
            "COREDUMP_PID": 123456789,
            "COREDUMP_PROC_MAPS": "mocked /proc/<pid>/maps",
            "COREDUMP_PROC_STATUS": "mocked /proc/<pid>/status",
            "COREDUMP_SIGNAL": 11,
            "COREDUMP_SIGNAL_NAME": "SIGSEGV",
            "COREDUMP_TIMESTAMP": datetime.datetime(
                2024, 2, 19, 12, 10, 42, tzinfo=datetime.timezone.utc
            ),
            "COREDUMP_UID": 0,
        }
        process_crash_mock.side_effect = lambda report, *_args: report

        report = apport_binary.process_crash_from_systemd_coredump("3-12345-7")

        get_systemd_coredump_mock.assert_called_once_with("3-12345-7")
        process_crash_mock.assert_called_once()
        self.assertEqual(report.pid, 123456789)
        self.assertEqual(
            dict(report),
            {
                "Date": "Mon Feb 19 12:10:42 2024",
                "CoreDump": b"(\xb5/\xfd$\x0ca\x00\x00mocked core\nG\xfe\xe0\x10",
                "ExecutablePath": sys.executable,
                "ExecutableTimestamp": str(int(os.stat(sys.executable).st_mtime)),
                "ProblemType": "Crash",
                "ProcCmdline": "python3",
                "ProcCwd": "/",
                "ProcEnviron": "SHELL=/bin/bash",
                "ProcMaps": "mocked /proc/<pid>/maps",
                "ProcStatus": "mocked /proc/<pid>/status",
                "Signal": "11",
                "SignalName": "SIGSEGV",
            },
        )



"""Test test helper functions. Test inception for the win!"""

# pylint: disable=missing-class-docstring,missing-function-docstring

import unittest

from tests.helper import get_init_system, wrap_object


class Multiply:  # pylint: disable=too-few-public-methods
    """Test class for wrap_object test cases."""

    def __init__(self, multiplier):
        self.multiplier = multiplier

    def multiply(self, factor: int) -> int:
        return factor * self.multiplier


class TestTestHelper(unittest.TestCase):
    def test_get_init_systemd(self):
        open_mock = unittest.mock.mock_open(read_data="systemd\n")
        with unittest.mock.patch("builtins.open", open_mock):
            self.assertEqual(get_init_system(), "systemd")
        open_mock.assert_called_once_with("/proc/1/comm", encoding="utf-8")

    def test_wrap_object_with_statement(self):
        with wrap_object(Multiply, "__init__") as mock:
            multiply = Multiply(7)
            self.assertEqual(multiply.multiply(6), 42)
        mock.assert_called_once_with(7)



"""Unit tests for the parse_segv general hook."""

import unittest

from problem_report import ProblemReport
from tests.helper import import_module_from_file
from tests.paths import get_data_directory

parse_segv = import_module_from_file(
    get_data_directory() / "general-hooks" / "parse_segv.py"
)

# Default global registers, maps, and disassembly for testing
REGS = """\
eax            0xffffffff -1
ecx            0xbfc6af40   -1077498048
edx            0x1  1
ebx            0x26eff4 2551796
esp            0xbfc6af24   0xbfc6af24
ebp            0xbfc6af28   0xbfc6af28
esi            0x826bb60    136756064
edi            0x8083480    134755456
eip            0x808354e    0x808354e <main+14>
eflags         0x200286 [ PF SF IF ID ]
cs             0x73 115
ss             0x7b 123
ds             0x7b 123
es             0x7b 123
fs             0x4  4
gs             0x33 51
"""
REGS64 = """\
rax            0xffffffffffffffff   -1
rbx            0x26eff4 2551796
rcx            0xffffffffffffffff   -1
rdx            0xffffffffff600180   -10485376
rsi            0x0  0
rdi            0x7fffffffe3b0   140737488348080
rbp            0x0  0x0
rsp            0x0000bfc6af24   0x0000bfc6af24
r8             0x0  0
r9             0x0  0
r10            0x7fffffffe140   140737488347456
r11            0x246    582
r12            0x7fffffffe400   140737488348160
r13            0x7fffffffe468   140737488348264
r14            0x1  1
r15            0x7fffffffe460   140737488348256
rip            0x7ffff790be10   0x7ffff790be10 <nanosleep+16>
eflags         0x246    [ PF ZF IF ]
cs             0x33 51
ss             0x2b 43
ds             0x0  0
es             0x0  0
fs             0x0  0
gs             0x0  0
fctrl          0x37f    895
fstat          0x0  0
ftag           0xffff   65535
fiseg          0x0  0
fioff          0x40303a 4206650
foseg          0x0  0
fooff          0x0  0
fop            0x5d8    1496
mxcsr          0x1f80   [ IM DM ZM OM UM PM ]
"""
MAPS = """\
00110000-0026c000 r-xp 00000000 08:06 375131     /lib/tls/i686/cmov/libc-2.9.so
0026c000-0026d000 ---p 0015c000 08:06 375131     /lib/tls/i686/cmov/libc-2.9.so
0026d000-0026f000 r--p 0015c000 08:06 375131     /lib/tls/i686/cmov/libc-2.9.so
0026f000-00270000 rw-p 0015e000 08:06 375131     /lib/tls/i686/cmov/libc-2.9.so
00270000-00273000 rw-p 00000000 00:00 0
002c1000-002e5000 r-xp 00000000 08:06 375135     /lib/tls/i686/cmov/libm-2.9.so
002e5000-002e6000 r--p 00023000 08:06 375135     /lib/tls/i686/cmov/libm-2.9.so
002e6000-002e7000 rw-p 00024000 08:06 375135     /lib/tls/i686/cmov/libm-2.9.so
00318000-00334000 r-xp 00000000 08:06 977846     /lib/ld-2.9.so
00334000-00335000 r--p 0001b000 08:06 977846     /lib/ld-2.9.so
00335000-00336000 rw-p 0001c000 08:06 977846     /lib/ld-2.9.so
0056e000-005a1000 r-xp 00000000 08:06 65575      /lib/libncurses.so.5.7
005a1000-005a3000 r--p 00033000 08:06 65575      /lib/libncurses.so.5.7
005a3000-005a4000 rw-p 00035000 08:06 65575      /lib/libncurses.so.5.7
00b67000-00b68000 r-xp 00000000 00:00 0          [vdso]
00bb6000-00bcb000 r-xp 00000000 08:06 375202 \
    /lib/tls/i686/cmov/libpthread-2.9.so
00bcb000-00bcc000 r--p 00014000 08:06 375202 \
    /lib/tls/i686/cmov/libpthread-2.9.so
00bcc000-00bcd000 rw-p 00015000 08:06 375202 \
    /lib/tls/i686/cmov/libpthread-2.9.so
00bcd000-00bcf000 rw-p 00000000 00:00 0
00beb000-00bed000 r-xp 00000000 08:06 375134 \
    /lib/tls/i686/cmov/libdl-2.9.so
00bed000-00bee000 r--p 00001000 08:06 375134 \
    /lib/tls/i686/cmov/libdl-2.9.so
00bee000-00bef000 rw-p 00002000 08:06 375134 \
    /lib/tls/i686/cmov/libdl-2.9.so
00c56000-00c7a000 r-xp 00000000 08:06 1140420    /usr/lib/libexpat.so.1.5.2
00c7a000-00c7c000 r--p 00023000 08:06 1140420    /usr/lib/libexpat.so.1.5.2
00c7c000-00c7d000 rw-p 00025000 08:06 1140420    /usr/lib/libexpat.so.1.5.2
00dce000-00dfa000 r-xp 00000000 08:06 65612      /lib/libreadline.so.5.2
00dfa000-00dfb000 ---p 0002c000 08:06 65612      /lib/libreadline.so.5.2
00dfb000-00dfc000 r--p 0002c000 08:06 65612      /lib/libreadline.so.5.2
00dfc000-00dff000 rw-p 0002d000 08:06 65612      /lib/libreadline.so.5.2
00dff000-00e00000 rw-p 00000000 00:00 0
08048000-0831c000 r-xp 00000000 08:06 1140349    /usr/bin/gdb
0831c000-0831d000 r--p 002d3000 08:06 1140349    /usr/bin/gdb
0831d000-08325000 rw-p 002d4000 08:06 1140349    /usr/bin/gdb
08325000-0833f000 rw-p 00000000 00:00 0
b8077000-b807a000 rw-p 00000000 00:00 0
b8096000-b8098000 rw-p 00000000 00:00 0
bfc57000-bfc6c000 rw-p 00000000 00:00 0          [stack]
"""
DISASM = """\
0x08083540 <main+0>:    lea    0x4(%esp),%ecx
0x08083544 <main+4>:    and    $0xfffffff0,%esp
0x08083547 <main+7>:    pushl  -0x4(%ecx)
0x0808354a <main+10>:   push   %ebp
0x0808354b <main+11>:   mov    %esp,%ebp
0x0808354d <main+13>:   push   %ecx
0x0808354e <main+14>:   sub    $0x14,%esp
0x08083551 <main+17>:   mov    (%ecx),%eax
0x08083553 <main+19>:   mov    0x4(%ecx),%edx
0x08083556 <main+22>:   lea    -0x14(%ebp),%ecx
0x08083559 <main+25>:   movl   $0x0,-0xc(%ebp)
0x08083560 <main+32>:   movl   $0x826bc68,-0x8(%ebp)
0x08083567 <main+39>:   mov    %eax,-0x14(%ebp)
0x0808356a <main+42>:   mov    %edx,-0x10(%ebp)
0x0808356d <main+45>:   mov    %ecx,(%esp)
0x08083570 <main+48>:   call   0x8083580 <gdb_main>
0x08083575 <main+53>:   add    $0x14,%esp
0x08083578 <main+56>:   pop    %ecx
0x08083579 <main+57>:   pop    %ebp
0x0808357a <main+58>:   lea    -0x4(%ecx),%esp
0x0808357d <main+61>:   ret
"""


class TestHookParseSegv(unittest.TestCase):
    """Test Segfault Parser."""

    def test_add_info_without_segv_reason(self) -> None:
        """Test add_info() with no SegvReason in result."""
        report = ProblemReport()
        report["Signal"] = "11"
        report["Architecture"] = "amd64"
        report["Disassembly"] = DISASM
        report["ProcMaps"] = MAPS
        report["Registers"] = REGS64

        parse_segv.add_info(report)

        self.assertEqual(
            report.get("SegvAnalysis"),
            "Segfault happened at: 0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"
            "PC (0x08083540) ok\n"
            "insn (lea) does not access VMA\n"
            "SP (0xbfc6af24) ok\n"
            "Reason could not be automatically determined.",
        )
        self.assertNotIn("SegvReason", report)

    def test_invalid_00_registers(self):
        """Require valid registers."""
        regs = "a 0x10\nb !!!\n"
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, "", "")
        try:
            segv = parse_segv.ParseSegv(regs, "", "")
        except ValueError as error:
            self.assertIn("invalid literal for int()", str(error))

        regs = "a 0x10"
        disasm = "0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.regs["a"], 0x10, segv)

        segv.regs = None
        self.assertRaises(ValueError, segv.parse_disassembly, "")

    def test_invalid_01_disassembly(self):
        # TODO: Split into separate test cases
        # pylint: disable=too-many-statements
        """Require valid disassembly."""
        regs = "a 0x10"

        disasm = ""
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, disasm, "")

        disasm = "Dump ..."
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, disasm, "")

        disasm = "Dump ...\nmonkey"
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, disasm, "")

        disasm = "monkey"
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, disasm, "")

        disasm = "0x1111111111: Cannot access memory at address 0x1111111111\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x1111111111, segv.pc)
        self.assertIsNone(segv.insn, segv.insn)
        self.assertIsNone(segv.src, segv.src)
        self.assertIsNone(segv.dest, segv.dest)

        disasm = "0x2111111111: \n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x2111111111, segv.pc)
        self.assertIsNone(segv.insn, segv.insn)
        self.assertIsNone(segv.src, segv.src)
        self.assertIsNone(segv.dest, segv.dest)

        disasm = "0x8069ff0 <fopen@plt+132220>: cmpb   $0x0,(%eax,%ebx,1)\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x8069FF0, segv.pc)
        self.assertEqual(segv.insn, "cmpb", segv.insn)
        self.assertEqual(segv.src, "$0x0", segv.src)
        self.assertEqual(segv.dest, "(%eax,%ebx,1)", segv.dest)

        disasm = "0xb765bb48 <_XSend+440>:  call   *0x40(%edi)\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0xB765BB48, segv.pc)
        self.assertEqual(segv.insn, "call", segv.insn)
        self.assertEqual(segv.src, "*0x40(%edi)", segv.src)
        self.assertIsNone(segv.dest, segv.dest)

        disasm = "0xb7aae5a0:   call   0xb7a805af <_Unwind_Find_FDE@plt+111>\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0xB7AAE5A0, segv.pc)
        self.assertEqual(segv.insn, "call", segv.insn)
        self.assertEqual(segv.src, "0xb7a805af", segv.src)
        self.assertIsNone(segv.dest, segv.dest)

        disasm = "0x09083540:    mov    0x4(%esp),%es:%ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x09083540, segv.pc)
        self.assertEqual(segv.insn, "mov", segv.insn)
        self.assertEqual(segv.src, "0x4(%esp)", segv.src)
        self.assertEqual(segv.dest, "%es:%ecx", segv.dest)

        disasm = "0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x08083540, segv.pc)
        self.assertEqual(segv.insn, "lea", segv.insn)
        self.assertEqual(segv.src, "0x4(%esp)", segv.src)
        self.assertEqual(segv.dest, "%ecx", segv.dest)

        disasm = (
            "0x404127 <exo_mount_hal_device_mount+167>:\n"
            "    repz cmpsb %es:(%rdi),%ds:(%rsi)\n"
        )
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x0404127, segv.pc)
        self.assertEqual(segv.insn, "repz cmpsb", segv.insn)
        self.assertEqual(segv.src, "%es:(%rdi)", segv.src)
        self.assertEqual(segv.dest, "%ds:(%rsi)", segv.dest)

        disasm = "0xb031765a <hufftab16+570>: add    0x3430433,%eax"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0xB031765A, segv.pc)
        self.assertEqual(segv.insn, "add", segv.insn)
        self.assertEqual(segv.src, "0x3430433", segv.src)
        self.assertEqual(segv.dest, "%eax", segv.dest)

        disasm = "Dump ...\n0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x08083540, segv.pc)
        self.assertEqual(segv.insn, "lea", segv.insn)
        self.assertEqual(segv.src, "0x4(%esp)", segv.src)
        self.assertEqual(segv.dest, "%ecx", segv.dest)

        disasm = "0x08083550 <main+0>:    nop\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x08083550, segv.pc)
        self.assertEqual(segv.insn, "nop", segv.insn)
        self.assertIsNone(segv.src, segv.src)
        self.assertIsNone(segv.dest, segv.dest)

        regs = "esp 0x444"
        disasm = "0x08083560 <main+0>:    push %ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x08083560, segv.pc)
        self.assertEqual(segv.insn, "push", segv.insn)
        self.assertEqual(segv.src, "%ecx", segv.src)
        self.assertEqual(segv.dest, "(%esp)", segv.dest)

        # GDB 7.1
        regs = "esp 0x444"
        disasm = "=> 0x08083560 <main+0>:    push %ecx\n"
        segv = parse_segv.ParseSegv(regs, disasm, "")
        self.assertEqual(segv.pc, 0x08083560, segv.pc)
        self.assertEqual(segv.insn, "push", segv.insn)
        self.assertEqual(segv.src, "%ecx", segv.src)
        self.assertEqual(segv.dest, "(%esp)", segv.dest)

    def test_ioport_operation(self):
        """I/O port violations."""
        regs = "rax            0x3  3"
        disasm = (
            "0x4087f1 <snd_pcm_hw_params_set_channels_near@plt+19345>:\n"
            "    out    %al,$0xb3\n"
        )
        maps = (
            "00400000-00412000 r-xp 00000000 08:04 10371157                   "
            "        /usr/sbin/pommed\n"
            "00611000-00614000 rw-p 00011000 08:04 10371157                   "
            "        /usr/sbin/pommed\n"
            "00614000-00635000 rw-p 00614000 00:00 0                          "
            "        [heap]\n"
        )
        segv = parse_segv.ParseSegv(regs, disasm, maps)
        self.assertEqual(segv.pc, 0x4087F1, segv.pc)
        self.assertEqual(segv.insn, "out", segv.insn)
        self.assertEqual(segv.src, "%al", segv.src)
        self.assertEqual(segv.dest, "$0xb3", segv.dest)

        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertTrue("disallowed I/O port operation on port 3" in reason, reason)

    def test_invalid_02_maps(self):
        """Require valid maps."""
        regs = "a 0x10"
        disasm = "Dump ...\n0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"

        maps = "asdlkfjaadf"
        self.assertRaises(ValueError, parse_segv.ParseSegv, regs, disasm, maps)

        maps = (
            "005a3000-005a4000 rw-p 00035000 08:06 65575  "
            "    /lib/libncurses.so.5.7\n"
            "00b67000-00b68000 r-xp 00000000 00:00 0          [vdso]\n"
            "00c67000-00c68000 r--p 00000000 00:00 0 "
        )
        segv = parse_segv.ParseSegv(regs, disasm, maps)
        self.assertEqual(segv.maps[0]["start"], 0x005A3000, segv)
        self.assertEqual(segv.maps[0]["end"], 0x005A4000, segv)
        self.assertEqual(segv.maps[0]["perms"], "rw-p", segv)
        self.assertEqual(segv.maps[0]["name"], "/lib/libncurses.so.5.7", segv)

        self.assertEqual(segv.maps[1]["start"], 0x00B67000, segv)
        self.assertEqual(segv.maps[1]["end"], 0x00B68000, segv)
        self.assertEqual(segv.maps[1]["perms"], "r-xp", segv)
        self.assertEqual(segv.maps[1]["name"], "[vdso]", segv)

        self.assertEqual(segv.maps[2]["start"], 0x00C67000, segv)
        self.assertEqual(segv.maps[2]["end"], 0x00C68000, segv)
        self.assertEqual(segv.maps[2]["perms"], "r--p", segv)
        self.assertIsNone(segv.maps[2]["name"], segv)

    def test_debug(self):
        """Debug mode works."""
        regs = "a 0x10"
        disasm = "Dump ...\n0x08083540 <main+0>:    lea    0x4(%esp),%ecx\n"
        maps = (
            "005a3000-005a4000 rw-p 00035000 08:06 65575  "
            "    /lib/libncurses.so.5.7\n"
            "00b67000-00b68000 r-xp 00000000 00:00 0          [vdso]\n"
            "00c67000-00c68000 r--p 00000000 00:00 0 "
        )
        with self.assertLogs(level="DEBUG"):
            segv = parse_segv.ParseSegv(regs, disasm, maps)
        self.assertTrue(segv is not None, segv)

    def test_register_values(self):
        """Sub-register parsing."""
        disasm = """0x08083540 <main+0>:    mov    $1,%ecx"""
        segv = parse_segv.ParseSegv(REGS64, disasm, "")

        val = segv.register_value("%rdx")
        self.assertEqual(val, 0xFFFFFFFFFF600180, hex(val))
        val = segv.register_value("%edx")
        self.assertEqual(val, 0xFF600180, hex(val))
        val = segv.register_value("%dx")
        self.assertEqual(val, 0x0180, hex(val))
        val = segv.register_value("%dl")
        self.assertEqual(val, 0x80, hex(val))

    def test_segv_unknown(self):
        """Handle unknown segfaults."""
        disasm = """0x08083540 <main+0>:    mov    $1,%ecx"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertFalse(understood, details)

        # Verify calculations
        self.assertEqual(segv.calculate_arg("(%ecx)"), 0xBFC6AF40, segv.regs["ecx"])
        self.assertEqual(segv.calculate_arg("0x10(%ecx)"), 0xBFC6AF50, segv.regs["ecx"])
        self.assertEqual(
            segv.calculate_arg("-0x20(%ecx)"), 0xBFC6AF20, segv.regs["ecx"]
        )
        self.assertEqual(segv.calculate_arg("%fs:(%ecx)"), 0xBFC6AF44, segv.regs["ecx"])
        self.assertEqual(segv.calculate_arg("0x3404403"), 0x3404403, "0x3404403")
        self.assertEqual(segv.calculate_arg("*0x40(%edi)"), 0x80834C0, segv.regs["edi"])
        self.assertEqual(
            segv.calculate_arg("(%edx,%ebx,1)"), 0x26EFF5, segv.regs["ebx"]
        )
        self.assertEqual(
            segv.calculate_arg("(%eax,%ebx,1)"), 0x26EFF3, segv.regs["ebx"]
        )
        self.assertEqual(
            segv.calculate_arg("0x10(,%ebx,1)"), 0x26F004, segv.regs["ebx"]
        )

        # Again, but 64bit
        disasm = """0x08083540 <main+0>:    mov    $1,%rcx"""
        segv = parse_segv.ParseSegv(REGS64, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertFalse(understood, details)

        self.assertEqual(
            segv.calculate_arg("(%rax,%rbx,1)"), 0x26EFF3, segv.regs["rbx"]
        )

    def test_segv_pc_missing(self):
        """Handle PC in missing VMA."""
        disasm = """0x00083540 <main+0>:    lea    0x4(%esp),%ecx"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn("PC (0x00083540) not located in a known VMA region", details)
        self.assertIn("executing unknown VMA", reason)

        disasm = """0x00083544:"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn("PC (0x00083544) not located in a known VMA region", details)
        self.assertIn("executing unknown VMA", reason)

    def test_segv_pc_null(self):
        """Handle PC in NULL VMA."""
        disasm = """0x00000540 <main+0>:    lea    0x4(%esp),%ecx"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn("PC (0x00000540) not located in a known VMA region", details)
        self.assertIn("executing NULL VMA", reason)

    def test_segv_pc_nx_writable(self):
        """Handle PC in writable NX VMA."""
        disasm = """0x005a3000 <main+0>:    lea    0x4(%esp),%ecx"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn("PC (0x005a3000) in non-executable VMA region:", details)
        self.assertIn("executing writable VMA /lib/libncurses.so.5.7", reason)

    def test_segv_pc_nx_unwritable(self):
        """Handle PC in non-writable NX VMA."""
        disasm = """0x00dfb000 <main+0>:    lea    0x4(%esp),%ecx"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn("PC (0x00dfb000) in non-executable VMA region:", details)
        self.assertIn("executing non-writable VMA /lib/libreadline.so.5.2", reason)

    def test_segv_src_missing(self):
        """Handle source in missing VMA."""
        reg = f"{REGS}ecx            0x0006af24   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"

        # Valid crash
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'source "-0x4(%ecx)" (0x0006af20) not located in a known VMA region',
            details,
        )
        self.assertIn("reading unknown VMA", reason)

        # Valid crash
        disasm = "0x08083547 <main+7>:    callq  *%ecx"
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'source "*%ecx" (0x0006af24) not located in a known VMA region', details
        )
        self.assertIn("reading unknown VMA", reason)

    def test_segv_src_null(self):
        """Handle source in NULL VMA."""
        reg = f"{REGS}ecx            0x00000024   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"

        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'source "-0x4(%ecx)" (0x00000020) not located in a known VMA region',
            details,
        )
        self.assertIn("reading NULL VMA", reason)

    def test_segv_src_not_readable(self):
        """Handle source not in readable VMA."""
        reg = f"{REGS}ecx            0x0026c080   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'source "-0x4(%ecx)" (0x0026c07c) in non-readable VMA region:', details
        )
        self.assertIn("reading VMA /lib/tls/i686/cmov/libc-2.9.so", reason)
        self.assertNotIn("Stack memory exhausted", details)
        self.assertNotIn("Stack pointer not within stack segment", details)

    def test_segv_dest_missing(self):
        """Handle destination in missing VMA."""
        reg = f"{REGS}esp            0x0006af24   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"

        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'destination "(%esp)" (0x0006af24) not located in a known VMA region',
            details,
        )
        self.assertIn("writing unknown VMA", reason)

    def test_segv_dest_null(self):
        """Handle destination in NULL VMA."""
        reg = f"{REGS}esp            0x00000024   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"

        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'destination "(%esp)" (0x00000024) not located in a known VMA region',
            details,
        )
        self.assertIn("writing NULL VMA", reason)

    def test_segv_dest_not_writable(self):
        """Handle destination not in writable VMA."""
        reg = f"{REGS}esp            0x08048080   0xbfc6af24"
        disasm = "0x08083547 <main+7>:    pushl  -0x4(%ecx)"
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, reason, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'destination "(%esp)" (0x08048080) in non-writable VMA region:', details
        )
        self.assertIn("writing VMA /usr/bin/gdb", reason)

    def test_segv_crackful_disasm(self):
        """Reject insane disassemblies."""
        disasm = "0x08083547 <main+7>:    pushl  -0x4(blah)"
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        self.assertRaises(ValueError, segv.report)

        disasm = "0x08083547 <main+7>:    pushl  -04(%ecx)"
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        self.assertRaises(ValueError, segv.report)

    def test_segv_stack_failure(self):
        """Handle walking off the stack."""
        # Triggered via "push"
        reg = f"{REGS}esp            0xbfc56ff0   0xbfc56ff0"
        disasm = "0x08083547 <main+7>:    push  %eax"
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'destination "(%esp)" (0xbfc56ff0) not located'
            " in a known VMA region (needed writable region)!",
            details,
        )

        # Triggered via "call"
        reg = f"{REGS}esp            0xbfc56fff   0xbfc56fff"
        disasm = "0x08083547 <main+7>:    callq  0x08083540"
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            'destination "(%esp)" (0xbfc56fff) not located'
            " in a known VMA region (needed writable region)!",
            details,
        )
        self.assertIn("Stack memory exhausted", details)

        # Triggered via unknown reason
        reg = f"{REGS}esp            0xdfc56000   0xdfc56000"
        disasm = """0x08083540 <main+0>:    mov    $1,%rcx"""
        segv = parse_segv.ParseSegv(reg, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertTrue(understood, details)
        self.assertIn(
            "SP (0xdfc56000) not located in a known VMA region"
            " (needed readable region)!",
            details,
        )
        self.assertIn("Stack pointer not within stack segment", details)

    def test_segv_stack_kernel_segfault(self):
        """Handle unknown segfaults in kernel."""
        # Crash in valid code path
        disasm = """0x0056e010: ret"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertFalse(understood, details)
        self.assertIn("Reason could not be automatically determined.", details)
        self.assertNotIn("(Unhandled exception in kernel code?)", details)

        # Crash from kernel code path
        disasm = """0x00b67422 <__kernel_vsyscall+2>: ret"""
        segv = parse_segv.ParseSegv(REGS, disasm, MAPS)
        understood, _, details = segv.report()
        self.assertFalse(understood, details)
        self.assertIn(
            "Reason could not be automatically determined."
            " (Unhandled exception in kernel code?)",
            details,
        )



"""Integration tests for the apport.fileutils module."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import os
import pwd
import shutil
import sys
import tempfile
import time
import unittest
import unittest.mock

import apport.fileutils
import apport.packaging
import problem_report


class T(unittest.TestCase):
    # pylint: disable=protected-access
    """Integration tests for the apport.fileutils module."""

    def setUp(self):
        self.orig_core_dir = apport.fileutils.core_dir
        apport.fileutils.core_dir = tempfile.mkdtemp()
        self.orig_report_dir = apport.fileutils.report_dir
        apport.fileutils.report_dir = tempfile.mkdtemp()
        self.orig_config_file = apport.fileutils._config_file

    def tearDown(self):
        shutil.rmtree(apport.fileutils.core_dir)
        apport.fileutils.core_dir = self.orig_core_dir
        shutil.rmtree(apport.fileutils.report_dir)
        apport.fileutils.report_dir = self.orig_report_dir
        self.orig_report_dir = None
        apport.fileutils._config_file = self.orig_config_file

    @staticmethod
    def _create_reports(create_inaccessible=False):
        """Create some test reports."""
        r1 = os.path.join(apport.fileutils.report_dir, "rep1.crash")
        r2 = os.path.join(apport.fileutils.report_dir, "rep2.crash")

        with open(r1, "w", encoding="utf-8") as fd:
            fd.write("report 1")
        with open(r2, "w", encoding="utf-8") as fd:
            fd.write("report 2")
        os.chmod(r1, 0o600)
        os.chmod(r2, 0o600)
        if create_inaccessible:
            ri = os.path.join(apport.fileutils.report_dir, "inaccessible.crash")
            with open(ri, "w", encoding="utf-8") as fd:
                fd.write("inaccessible")
            os.chmod(ri, 0)
            return [r1, r2, ri]
        return [r1, r2]

    def test_find_package_desktopfile(self):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches
        """find_package_desktopfile()"""
        # package without any .desktop file
        nodesktop = "bash"
        assert (
            len(
                [
                    f
                    for f in apport.packaging.get_files(nodesktop)
                    if f.endswith(".desktop")
                ]
            )
            == 0
        )

        # find a package with one, a package with multiple .desktop files, and
        # a package with a NoDisplay .desktop file
        onedesktop = None
        multidesktop = None
        nodisplay = None
        found_some = False
        for d in sorted(os.listdir("/usr/share/applications/")):
            if not d.endswith(".desktop"):
                continue
            path = os.path.join("/usr/share/applications/", d)
            pkg = apport.packaging.get_file_package(path)
            if pkg is None:
                continue
            found_some = True

            display_num = 0
            no_display_num = 0
            for desktop_file in apport.packaging.get_files(pkg):
                if not desktop_file.endswith(".desktop"):
                    continue
                if desktop_file.startswith("/usr/share/mimelnk"):
                    continue
                with open(desktop_file, "rb") as desktop_file:
                    if b"NoDisplay=true" in desktop_file.read():
                        no_display_num += 1
                    else:
                        display_num += 1

            if not nodisplay and display_num == 0 and no_display_num == 1:
                nodisplay = pkg
            elif not onedesktop and display_num == 1:
                onedesktop = pkg
            elif not multidesktop and display_num > 1:
                multidesktop = pkg

            if onedesktop and multidesktop and nodisplay:
                break

        self.assertTrue(found_some)

        if nodesktop:
            self.assertEqual(
                apport.fileutils.find_package_desktopfile(nodesktop),
                None,
                f"no-desktop package {nodesktop}",
            )
        if multidesktop:
            self.assertEqual(
                apport.fileutils.find_package_desktopfile(multidesktop),
                None,
                f"multi-desktop package {multidesktop}",
            )
        if onedesktop:
            d = apport.fileutils.find_package_desktopfile(onedesktop)
            self.assertIsNotNone(d, f"one-desktop package {onedesktop}")
            self.assertTrue(os.path.exists(d))
            self.assertTrue(d.endswith(".desktop"))
        if nodisplay:
            self.assertEqual(
                apport.fileutils.find_package_desktopfile(nodisplay),
                None,
                f"NoDisplay package {nodisplay}",
            )

    def test_find_file_package(self):
        """find_file_package()"""
        self.assertEqual(apport.fileutils.find_file_package("/bin/bash"), "bash")
        self.assertEqual(apport.fileutils.find_file_package("/bin/cat"), "coreutils")
        self.assertIsNone(apport.fileutils.find_file_package("/nonexisting"))

    def test_seen(self):
        """get_new_reports() and seen_report()"""
        self.assertEqual(apport.fileutils.get_new_reports(), [])
        if os.getuid() == 0:
            tr = self._create_reports(True)
        else:
            tr = [r for r in self._create_reports(True) if "inaccessible" not in r]
        self.assertEqual(set(apport.fileutils.get_new_reports()), set(tr))

        # now mark them as seen and check again
        nr = set(tr)
        for r in tr:
            self.assertEqual(apport.fileutils.seen_report(r), False)
            nr.remove(r)
            apport.fileutils.mark_report_seen(r)
            self.assertEqual(apport.fileutils.seen_report(r), True)
            self.assertEqual(set(apport.fileutils.get_new_reports()), nr)

    def test_mark_hanging_process(self):
        """mark_hanging_process()"""
        pr = problem_report.ProblemReport()
        pr["ExecutablePath"] = "/bin/bash"
        apport.fileutils.mark_hanging_process(pr, "1")
        uid = str(os.getuid())
        base = f"_bin_bash.{uid}.1.hanging"
        expected = os.path.join(apport.fileutils.report_dir, base)
        self.assertTrue(os.path.exists(expected))

    def test_mark_report_upload(self):
        """mark_report_upload()"""
        report = os.path.join(apport.fileutils.report_dir, "report.crash")
        apport.fileutils.mark_report_upload(report)
        expected = os.path.join(apport.fileutils.report_dir, "report.upload")
        self.assertTrue(os.path.exists(expected))

    def test_mark_2nd_report_upload(self):
        """mark_report_upload() for a previously uploaded report"""
        upload = os.path.join(apport.fileutils.report_dir, "report.upload")
        with open(upload, "w", encoding="utf-8"):
            pass
        uploaded = os.path.join(apport.fileutils.report_dir, "report.uploaded")
        with open(uploaded, "w", encoding="utf-8"):
            pass
        time.sleep(1)
        report = os.path.join(apport.fileutils.report_dir, "report.crash")
        with open(report, "w", encoding="utf-8"):
            pass
        time.sleep(1)
        apport.fileutils.mark_report_upload(report)
        upload_st = os.stat(upload)
        report_st = os.stat(report)
        self.assertTrue(upload_st.st_mtime > report_st.st_mtime)

    def test_get_all_reports(self):
        """get_all_reports()"""
        self.assertEqual(apport.fileutils.get_all_reports(), [])
        if os.getuid() == 0:
            tr = self._create_reports(True)
        else:
            tr = [r for r in self._create_reports(True) if "inaccessible" not in r]
        self.assertEqual(set(apport.fileutils.get_all_reports()), set(tr))

        # now mark them as seen and check again
        for r in tr:
            apport.fileutils.mark_report_seen(r)

        self.assertEqual(set(apport.fileutils.get_all_reports()), set(tr))

    def test_get_system_reports(self):
        """get_all_system_reports() and get_new_system_reports()"""
        self.assertEqual(apport.fileutils.get_all_reports(), [])
        self.assertEqual(apport.fileutils.get_all_system_reports(), [])
        if os.getuid() == 0:
            tr = self._create_reports(True)
            self.assertEqual(set(apport.fileutils.get_all_system_reports()), set(tr))
            self.assertEqual(set(apport.fileutils.get_new_system_reports()), set(tr))

            # now mark them as seen and check again
            for r in tr:
                apport.fileutils.mark_report_seen(r)

            self.assertEqual(set(apport.fileutils.get_all_system_reports()), set(tr))
            self.assertEqual(set(apport.fileutils.get_new_system_reports()), set([]))
        else:
            tr = [r for r in self._create_reports(True) if "inaccessible" not in r]
            self.assertEqual(set(apport.fileutils.get_all_system_reports()), set([]))
            self.assertEqual(set(apport.fileutils.get_new_system_reports()), set([]))

    @unittest.mock.patch.object(os, "stat")
    @unittest.mock.patch.object(pwd, "getpwuid")
    def test_get_system_reports_guest(self, getpwuid_mock, stat_mock):
        """get_all_system_reports() filters out reports from guest user"""
        self._create_reports()

        stat_mock.return_value.st_size = 1000
        stat_mock.return_value.st_uid = 123
        getpwuid_mock.return_value.pw_name = "guest_tmp987"
        self.assertEqual(apport.fileutils.get_all_system_reports(), [])

    def test_unwritable_report(self):
        """get_all_reports() and get_new_reports() for unwritable report"""
        self.assertEqual(apport.fileutils.get_all_reports(), [])
        self.assertEqual(apport.fileutils.get_all_system_reports(), [])

        r = os.path.join(apport.fileutils.report_dir, "unwritable.crash")
        with open(r, "w", encoding="utf-8") as fd:
            fd.write("unwritable")
        os.chmod(r, 0o444)

        if os.getuid() == 0:
            self.assertEqual(set(apport.fileutils.get_new_reports()), set([r]))
            self.assertEqual(set(apport.fileutils.get_all_reports()), set([r]))
        else:
            self.assertEqual(set(apport.fileutils.get_new_reports()), set())
            self.assertEqual(set(apport.fileutils.get_all_reports()), set())

    def test_delete_report(self):
        """delete_report()"""
        tr = self._create_reports()

        while tr:
            self.assertEqual(set(apport.fileutils.get_all_reports()), set(tr))
            apport.fileutils.delete_report(tr.pop())

    def test_make_report_file(self):
        """make_report_file()"""
        pr = problem_report.ProblemReport()
        self.assertRaises(ValueError, apport.fileutils.make_report_file, pr)

        pr["Package"] = "bash 1"
        with apport.fileutils.make_report_file(pr) as f:
            path = f.name
            self.assertTrue(
                path.startswith(f"{apport.fileutils.report_dir}/bash"), path
            )
            os.unlink(path)

        pr["ExecutablePath"] = "/bin/bash"
        with apport.fileutils.make_report_file(pr) as f:
            path = f.name
            self.assertTrue(
                path.startswith(f"{apport.fileutils.report_dir}/_bin_bash"), path
            )

        # file exists already, should fail now
        self.assertRaises(OSError, apport.fileutils.make_report_file, pr)

        # should still fail if it's a dangling symlink
        os.unlink(path)
        os.symlink(os.path.join(apport.fileutils.report_dir, "pwned"), path)
        self.assertRaises(OSError, apport.fileutils.make_report_file, pr)

    def test_check_files_md5(self):
        """check_files_md5()"""
        f1 = os.path.join(apport.fileutils.report_dir, "test 1.txt")
        f2 = os.path.join(apport.fileutils.report_dir, "test:2.txt")
        sumfile = os.path.join(apport.fileutils.report_dir, "sums.txt")
        with open(f1, "w", encoding="utf-8") as fd:
            fd.write("Some stuff")
        with open(f2, "w", encoding="utf-8") as fd:
            fd.write("More stuff")
        # use one relative and one absolute path in checksums file
        with open(sumfile, "w", encoding="utf-8") as fd:
            fd.write(
                f"""2e41290da2fa3f68bd3313174467e3b5  {f1[1:]}
f6423dfbc4faf022e58b4d3f5ff71a70  {f2}
"""
            )
        self.assertEqual(
            apport.fileutils.check_files_md5(sumfile), [], "correct md5sums"
        )

        with open(f1, "w", encoding="utf-8") as fd:
            fd.write("Some stuff!")
        self.assertEqual(
            apport.fileutils.check_files_md5(sumfile), [f1[1:]], "file 1 wrong"
        )
        with open(f2, "w", encoding="utf-8") as fd:
            fd.write("More stuff!")
        self.assertEqual(
            apport.fileutils.check_files_md5(sumfile),
            [f1[1:], f2],
            "files 1 and 2 wrong",
        )
        with open(f1, "w", encoding="utf-8") as fd:
            fd.write("Some stuff")
        self.assertEqual(
            apport.fileutils.check_files_md5(sumfile), [f2], "file 2 wrong"
        )

    def test_get_config(self):
        """get_config()"""
        # nonexisting
        apport.fileutils._config_file = "/nonexisting"
        self.assertIsNone(apport.fileutils.get_config("main", "foo"))
        self.assertEqual(apport.fileutils.get_config("main", "foo", "moo"), "moo")
        apport.fileutils._get_config_parser.cache_clear()

        # empty
        with tempfile.NamedTemporaryFile() as f:
            apport.fileutils._config_file = f.name
            self.assertIsNone(apport.fileutils.get_config("main", "foo"))
            self.assertEqual(apport.fileutils.get_config("main", "foo", "moo"), "moo")
            apport.fileutils._get_config_parser.cache_clear()

            # nonempty
            f.write(
                b"[main]\none=1\ntwo = TWO\nb1 = 1\nb2=False\n" b"[spethial]\none= 99\n"
            )
            f.flush()
            self.assertIsNone(apport.fileutils.get_config("main", "foo"))
            self.assertEqual(apport.fileutils.get_config("main", "foo", "moo"), "moo")
            self.assertEqual(apport.fileutils.get_config("main", "one"), "1")
            self.assertEqual(
                apport.fileutils.get_config("main", "one", default="moo"), "1"
            )
            self.assertEqual(apport.fileutils.get_config("main", "two"), "TWO")
            self.assertEqual(
                apport.fileutils.get_config("main", "b1", boolean=True), True
            )
            self.assertEqual(
                apport.fileutils.get_config("main", "b2", boolean=True), False
            )
            self.assertEqual(
                apport.fileutils.get_config("main", "b3", boolean=True), None
            )
            self.assertEqual(
                apport.fileutils.get_config("main", "b3", default=False, boolean=True),
                False,
            )
            self.assertEqual(apport.fileutils.get_config("spethial", "one"), "99")
            self.assertIsNone(apport.fileutils.get_config("spethial", "two"))
            self.assertEqual(
                apport.fileutils.get_config("spethial", "one", "moo"), "99"
            )
            self.assertEqual(
                apport.fileutils.get_config("spethial", "nope", "moo"), "moo"
            )
            apport.fileutils._get_config_parser.cache_clear()

            # interpolation
            f.write(b"[inter]\none=1\ntwo = TWO\ntest = %(two)s\n")
            f.flush()
            self.assertEqual(apport.fileutils.get_config("inter", "one"), "1")
            self.assertEqual(apport.fileutils.get_config("inter", "two"), "TWO")
            self.assertEqual(apport.fileutils.get_config("inter", "test"), "%(two)s")
            apport.fileutils._get_config_parser.cache_clear()

    def test_shared_libraries(self):
        """shared_libraries()"""
        libs = apport.fileutils.shared_libraries(sys.executable)
        self.assertGreater(len(libs), 3)
        self.assertIn("libc.so.6", libs)
        self.assertIn("libc.so.6", libs["libc.so.6"])
        self.assertTrue(os.path.exists(libs["libc.so.6"]))
        for library_name, library_path in libs.items():
            self.assertNotIn("vdso", library_name, libs)
            self.assertTrue(os.path.exists(library_path))

        self.assertEqual(apport.fileutils.shared_libraries("/non/existing"), {})
        self.assertEqual(apport.fileutils.shared_libraries("/etc"), {})
        self.assertEqual(apport.fileutils.shared_libraries("/etc/passwd"), {})

    def test_links_with_shared_library(self):
        """links_with_shared_library()"""
        self.assertTrue(
            apport.fileutils.links_with_shared_library(sys.executable, "libc")
        )
        self.assertTrue(
            apport.fileutils.links_with_shared_library(sys.executable, "libc.so.6")
        )
        self.assertFalse(
            apport.fileutils.links_with_shared_library(sys.executable, "libc.so.7")
        )
        self.assertFalse(
            apport.fileutils.links_with_shared_library(sys.executable, "libd")
        )
        self.assertFalse(
            apport.fileutils.links_with_shared_library("/non/existing", "libc")
        )
        self.assertFalse(apport.fileutils.links_with_shared_library("/etc", "libc"))
        self.assertFalse(
            apport.fileutils.links_with_shared_library("/etc/passwd", "libc")
        )

    def test_get_core_path(self):
        """get_core_path()"""
        boot_id = apport.fileutils.get_boot_id()

        # Basic test
        (core_name, core_path) = apport.fileutils.get_core_path(
            pid=123, exe="/usr/bin/test", uid=234, timestamp=222222
        )
        expected = f"core._usr_bin_test.234.{boot_id}.123.222222"
        expected_path = os.path.join(apport.fileutils.core_dir, expected)
        self.assertEqual(core_name, expected)
        self.assertEqual(core_path, expected_path)

        # Test dots in exe names
        (core_name, core_path) = apport.fileutils.get_core_path(
            pid=123, exe="/usr/bin/test.sh", uid=234, timestamp=222222
        )
        expected = f"core._usr_bin_test_sh.234.{boot_id}.123.222222"
        expected_path = os.path.join(apport.fileutils.core_dir, expected)
        self.assertEqual(core_name, expected)
        self.assertEqual(core_path, expected_path)

        # Test no exe name
        (core_name, core_path) = apport.fileutils.get_core_path(
            pid=123, exe=None, uid=234, timestamp=222222
        )
        expected = f"core.unknown.234.{boot_id}.123.222222"
        expected_path = os.path.join(apport.fileutils.core_dir, expected)
        self.assertEqual(core_name, expected)
        self.assertEqual(core_path, expected_path)

        # Test no uid
        (core_name, core_path) = apport.fileutils.get_core_path(
            pid=123, exe="/usr/bin/test", uid=None, timestamp=222222
        )
        expected = f"core._usr_bin_test.{str(os.getuid())}.{boot_id}.123.222222"
        expected_path = os.path.join(apport.fileutils.core_dir, expected)
        self.assertEqual(core_name, expected)
        self.assertEqual(core_path, expected_path)

    def test_clean_core_directory(self):
        """clean_core_directory()"""
        fake_uid = 5150
        extra_core_files = 4
        num_core_files = apport.fileutils.max_corefiles_per_uid + extra_core_files

        # Create some test files
        for x in range(num_core_files):
            core_path = apport.fileutils.get_core_path(
                pid=123 + x, exe="/usr/bin/test", uid=fake_uid, timestamp=222222 + x
            )[1]
            with open(core_path, "w", encoding="utf-8") as fd:
                fd.write("Some stuff")
                time.sleep(1)

        # Create a file with a different uid
        core_path = apport.fileutils.get_core_path(
            pid=231, exe="/usr/bin/test", uid=fake_uid + 1, timestamp=333333
        )[1]
        with open(core_path, "w", encoding="utf-8") as fd:
            fd.write("Some stuff")

        # Make sure we have the proper number of test files
        self.assertEqual(
            num_core_files, len(apport.fileutils.find_core_files_by_uid(fake_uid))
        )
        self.assertEqual(1, len(apport.fileutils.find_core_files_by_uid(fake_uid + 1)))

        # Clean out the directory
        apport.fileutils.clean_core_directory(fake_uid)

        # Make sure we have the proper number of test files. We should
        # have one less than max_corefiles_per_uid.
        self.assertEqual(
            apport.fileutils.max_corefiles_per_uid - 1,
            len(apport.fileutils.find_core_files_by_uid(fake_uid)),
        )
        self.assertEqual(1, len(apport.fileutils.find_core_files_by_uid(fake_uid + 1)))

        # Make sure we deleted the oldest ones
        for x in range(apport.fileutils.max_corefiles_per_uid - 1):
            offset = extra_core_files + x + 1
            core_path = apport.fileutils.get_core_path(
                pid=123 + offset,
                exe="/usr/bin/test",
                uid=fake_uid,
                timestamp=222222 + offset,
            )[1]
            self.assertTrue(os.path.exists(core_path))



"""System tests for data/apport."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import errno
import os
import resource
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import unittest

import apport.fileutils
from tests.helper import get_init_system, pidof, skip_if_command_is_missing
from tests.paths import (
    get_data_directory,
    is_local_source_directory,
    local_test_environment,
)


class T(unittest.TestCase):
    # pylint: disable=missing-class-docstring,missing-function-docstring
    # pylint: disable=protected-access
    TEST_EXECUTABLE = os.path.realpath("/bin/sleep")
    TEST_ARGS = ["86400"]

    @classmethod
    def setUpClass(cls):
        cls.orig_environ = os.environ.copy()
        os.environ |= local_test_environment()

        with open("/proc/sys/kernel/core_pattern", encoding="utf-8") as f:
            core_pattern = f.read().strip()
        if core_pattern[0] == "|":
            cls.apport_path = core_pattern[1:].split()[0]
        else:
            cls.apport_path = None
        if is_local_source_directory():
            cls.apport_path = get_data_directory() / "apport"

        cls.all_reports = apport.fileutils.get_all_reports()

        # ensure we don't inherit an ignored SIGQUIT
        signal.signal(signal.SIGQUIT, signal.SIG_DFL)

        orig_home = os.getenv("HOME")
        if orig_home is not None:
            del os.environ["HOME"]
        cls.ifpath = os.path.expanduser(apport.report._ignore_file)
        if orig_home is not None:
            os.environ["HOME"] = orig_home

        cls.orig_cwd = os.getcwd()
        cls.orig_core_dir = apport.fileutils.core_dir
        cls.orig_ignore_file = apport.report._ignore_file
        cls.orig_report_dir = apport.fileutils.report_dir

    @classmethod
    def tearDownClass(cls):
        os.environ.clear()
        os.environ.update(cls.orig_environ)
        apport.fileutils.core_dir = cls.orig_core_dir
        apport.report._ignore_file = cls.orig_ignore_file
        apport.fileutils.report_dir = cls.orig_report_dir
        os.chdir(cls.orig_cwd)

    def setUp(self):
        if self.apport_path is None:
            self.skipTest(
                "kernel crash dump helper is not active; please enable"
                " before running this test"
            )

        if self.all_reports:
            self.skipTest(
                "Please remove all crash reports from /var/crash/ for this"
                " test suite:\n  %s\n" % "\n  ".join(self.all_reports)
            )

        # use local report dir
        self.report_dir = tempfile.mkdtemp()
        os.environ["APPORT_REPORT_DIR"] = self.report_dir
        apport.fileutils.report_dir = self.report_dir

        self.workdir = tempfile.mkdtemp()

        apport.fileutils.core_dir = os.path.join(self.workdir, "coredump")
        os.mkdir(apport.fileutils.core_dir)
        os.environ["APPORT_COREDUMP_DIR"] = apport.fileutils.core_dir

        os.environ["APPORT_IGNORE_FILE"] = os.path.join(
            self.workdir, "apport-ignore.xml"
        )
        apport.report._ignore_file = os.environ["APPORT_IGNORE_FILE"]

        os.environ["APPORT_LOCK_FILE"] = os.path.join(self.workdir, "apport.lock")

        # move aside current ignore file
        if os.path.exists(self.ifpath):
            os.rename(self.ifpath, f"{self.ifpath}.apporttest")

        # do not write core files by default
        resource.setrlimit(resource.RLIMIT_CORE, (0, -1))

        # that's the place where to put core dumps, etc.
        os.chdir("/tmp")

        # expected report name for test executable report
        self.test_report = os.path.join(
            apport.fileutils.report_dir,
            f"{self.TEST_EXECUTABLE.replace('/', '_')}.{os.getuid()}.crash",
        )

        self.running_test_executables = pidof(self.TEST_EXECUTABLE)

    def tearDown(self):
        shutil.rmtree(self.report_dir)
        shutil.rmtree(self.workdir)

        # clean up our ignore file
        if os.path.exists(self.ifpath):
            os.unlink(self.ifpath)
        orig_ignore_file = f"{self.ifpath}.apporttest"
        if os.path.exists(orig_ignore_file):
            os.rename(orig_ignore_file, self.ifpath)

        # permit tests to leave behind test_report, but nothing else
        if os.path.exists(self.test_report):
            apport.fileutils.delete_report(self.test_report)
        unexpected_reports = apport.fileutils.get_all_reports()
        for r in unexpected_reports:
            apport.fileutils.delete_report(r)
        self.assertEqual(unexpected_reports, [])

    def test_limit_size(self):
        """Core dumps are capped on available memory size."""
        # determine how much data we have to pump into apport in order to make
        # sure that it will refuse the core dump
        r = apport.Report()
        with open("/proc/meminfo", "rb") as f:
            r.load(f)
        totalmb = int(r["MemFree"].split()[0]) + int(r["Cached"].split()[0])
        totalmb = int(totalmb / 1024)
        r = None

        test_proc = self.create_test_process()
        try:
            with subprocess.Popen(
                [
                    self.apport_path,
                    "-p",
                    str(test_proc.pid),
                    "-s",
                    "42",
                    "-c",
                    "0",
                    "-d",
                    "1",
                ],
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE,
            ) as app:
                # pipe an entire total memory size worth of spaces into it,
                # which must be bigger than the 'usable' memory size. apport
                # should digest that and the report should not have a core
                # dump; NB that this should error out with a SIGPIPE when
                # apport aborts reading from stdin
                onemb = b" " * 1048576
                while totalmb > 0:
                    if totalmb & 255 == 0:
                        # Print a dot every 256 MiB
                        sys.stderr.write(".")
                        sys.stderr.flush()
                    try:
                        app.stdin.write(onemb)
                    except OSError as error:
                        if error.errno == errno.EPIPE:
                            break
                        raise
                    totalmb -= 1
                err = app.communicate()[1]
            self.assertEqual(app.returncode, 0, err)
            onemb = None
        finally:
            test_proc.kill()
            test_proc.wait()

        reports = apport.fileutils.get_all_reports()
        self.assertEqual(reports, [self.test_report])

        pr = apport.Report()
        with open(reports[0], "rb") as f:
            pr.load(f)
        os.unlink(reports[0])

        self.assertEqual(pr["Signal"], "42")
        self.assertEqual(pr["ExecutablePath"], self.TEST_EXECUTABLE)
        self.assertNotIn("CoreDump", pr)
        # FIXME: sometimes this is empty!?
        if err:
            self.assertRegex(
                err.decode(),
                f"core dump exceeded.*dropped from .*"
                f"{os.path.basename(self.TEST_EXECUTABLE)}\\..*\\.crash",
            )

    @unittest.skipIf(
        get_init_system() != "systemd", "running init system is not systemd"
    )
    @skip_if_command_is_missing("systemd-run")
    @unittest.skipIf(os.geteuid() != 0, "this test needs to be run as root")
    def test_crash_system_slice(self):
        """Report generation for a protected process running in the system
        slice"""
        apport.fileutils.report_dir = self.orig_report_dir
        self.test_report = os.path.join(
            apport.fileutils.report_dir,
            f"{self.TEST_EXECUTABLE.replace('/', '_')}.{os.getuid()}.crash",
        )

        system_run = self.create_test_process(
            command="/usr/bin/systemd-run",
            args=[
                "-t",
                "-q",
                "--slice=system.slice",
                "-p",
                "ProtectSystem=true",
                self.TEST_EXECUTABLE,
            ]
            + self.TEST_ARGS,
        )
        try:
            pids = pidof(self.TEST_EXECUTABLE) - self.running_test_executables
            self.assertEqual(len(pids), 1)
            os.kill(pids.pop(), signal.SIGSEGV)

            self.wait_for_no_instance_running(self.TEST_EXECUTABLE)
            self.wait_for_apport_to_finish()
        finally:
            system_run.kill()
            system_run.wait()

        # check crash report
        with open("/var/log/apport.log", encoding="utf-8") as logfile:
            apport_log = logfile.read().strip()
        reports = apport.fileutils.get_all_reports()
        self.assertEqual(reports, [self.test_report], f"Apport log:\n{apport_log}")
        self.assertEqual(
            reports[0], f"/var/crash/{self.TEST_EXECUTABLE.replace('/', '_')}.0.crash"
        )

    def create_test_process(self, command=None, uid=None, args=None):
        """Spawn test executable.

        Wait until it is fully running, and return its process.
        """
        if command is None:
            command = self.TEST_EXECUTABLE
        if args is None:
            args = self.TEST_ARGS

        assert os.access(command, os.X_OK), f"{command} is not executable"

        env = os.environ.copy()
        # set UTF-8 environment variable, to check proper parsing in apport
        os.putenv("utf8trap", b"\xc3\xa0\xc3\xa4")
        # caller needs to call .wait(), pylint: disable=consider-using-with
        # False positive, see https://github.com/PyCQA/pylint/issues/7092
        process = subprocess.Popen(  # pylint: disable=unexpected-keyword-arg
            [command] + args, env=env, stdout=subprocess.DEVNULL, user=uid
        )

        # wait until child process has execv()ed properly
        while True:
            with open(f"/proc/{process.pid}/cmdline", encoding="utf-8") as f:
                cmdline = f.read()
            if "test_signal" in cmdline:
                time.sleep(0.1)
            else:
                break

        time.sleep(0.3)  # needs some more setup time
        return process

    def wait_for_apport_to_finish(self, timeout_sec=10.0):
        self.wait_for_no_instance_running(self.apport_path, timeout_sec)

    def wait_for_no_instance_running(self, program, timeout_sec=10.0):
        while timeout_sec > 0:
            if not pidof(program) - self.running_test_executables:
                break
            time.sleep(0.2)
            timeout_sec -= 0.2
        else:
            self.fail(f"Timeout exceeded, but {program} is still running.")
